@@ -33,76 +33,21 @@ class BuildSlaveTooOldError(Exception):
 class LatentBuildSlaveFailedToSubstantiate(Exception):
     pass
 
-# other exceptions
-class BuildbotNotRunningError(Exception):
-    pass
-
-class ParameterError(Exception):
-    pass
-
 class IChangeSource(Interface):
-    """Object which feeds Change objects to the changemaster. When files or
-    directories are changed and the version control system provides some
-    kind of notification, this object should turn it into a Change object
-    and pass it through::
-
-      self.changemaster.addChange(change)
     """
+    Service which feeds Change objects to the changemaster. When files or
+    directories are changed in version control, this object should represent
+    the changes as a change dictionary and call::
+
+      self.master.addChange(who=.., rev=.., ..)
+
+    See 'Writing Change Sources' in the manual for more information.
+    """
+    master = Attribute('master',
+            'Pointer to BuildMaster, automatically set when started.')
 
     def describe():
-        """Should return a string which briefly describes this source. This
-        string will be displayed in an HTML status page."""
-
-class IScheduler(Interface):
-    """I watch for Changes in the source tree and decide when to trigger
-    Builds. I create BuildSet objects and submit them to the BuildMaster. I
-    am a service, and the BuildMaster is always my parent.
-
-    @ivar properties: properties to be applied to all builds started by this
-    scheduler
-    @type properties: L<buildbot.process.properties.Properties>
-    """
-
-    def addChange(change):
-        """A Change has just been dispatched by one of the ChangeSources.
-        Each Scheduler will receive this Change. I may decide to start a
-        build as a result, or I might choose to ignore it."""
-
-    def listBuilderNames():
-        """Return a list of strings indicating the Builders that this
-        Scheduler might feed."""
-
-    def getPendingBuildTimes():
-        """Return a list of timestamps for any builds that are waiting in the
-        tree-stable-timer queue. This is only relevant for Change-based
-        schedulers, all others can just return an empty list."""
-        # TODO: it might be nice to make this into getPendingBuildSets, which
-        # would let someone subscribe to the buildset being finished.
-        # However, the Scheduler doesn't actually create the buildset until
-        # it gets submitted, so doing this would require some major rework.
-
-class IUpstreamScheduler(Interface):
-    """This marks an IScheduler as being eligible for use as the 'upstream='
-    argument to a buildbot.scheduler.Dependent instance."""
-
-    def subscribeToSuccessfulBuilds(target):
-        """Request that the target callbable be invoked after every
-        successful buildset. The target will be called with a single
-        argument: the SourceStamp used by the successful builds."""
-
-    def listBuilderNames():
-        """Return a list of strings indicating the Builders that this
-        Scheduler might feed."""
-
-class IDownstreamScheduler(Interface):
-    """This marks an IScheduler to be listening to other schedulers.
-    On reconfigs, these might get notified to check if their upstream
-    scheduler are stil the same."""
-
-    def checkUpstreamScheduler():
-        """Check if the upstream scheduler is still alive, and if not,
-        get a new upstream object from the master."""
-
+        """Return a string which briefly describes this source."""
 
 class ISourceStamp(Interface):
     """
@@ -162,10 +107,10 @@ class IStatus(Interface):
     """I am an object, obtainable from the buildmaster, which can provide
     status information."""
 
-    def getProjectName():
+    def getTitle():
         """Return the name of the project that this Buildbot is working
         for."""
-    def getProjectURL():
+    def getTitleURL():
         """Return the URL of this Buildbot's project."""
     def getBuildbotURL():
         """Return the URL of the top-most Buildbot status page, or None if
@@ -199,7 +144,11 @@ class IStatus(Interface):
         """Return the ISlaveStatus object for a given named buildslave."""
 
     def getBuildSets():
-        """Return a list of active (non-finished) IBuildSetStatus objects."""
+        """
+        Return a list of un-completed build sets.
+
+        @returns: list of L{IBuildSetStatus} implementations, via Deferred.
+        """
 
     def generateFinishedBuilds(builders=[], branches=[],
                                num_builds=None, finished_before=None,
@@ -254,13 +203,6 @@ class IBuildSetStatus(Interface):
     """I represent a set of Builds, each run on a separate Builder but all
     using the same source tree."""
 
-    def getSourceStamp():
-        """Return a SourceStamp object which can be used to re-create
-        the source tree that this build used.
-
-        This method will return None if the source information is no longer
-        available."""
-        pass
     def getReason():
         pass
     def getID():
@@ -273,18 +215,11 @@ class IBuildSetStatus(Interface):
         pass # not implemented
     def getBuilderNames():
         """Return a list of the names of all Builders on which this set will
-        do builds."""
-    def getBuildRequests():
-        """Return a list of IBuildRequestStatus objects that represent my
-        component Builds. This list might correspond to the Builders named by
-        getBuilderNames(), but if builder categories are used, or 'Builder
-        Aliases' are implemented, then they may not."""
+        do builds.
+        
+        @returns: list of names via Deferred"""
     def isFinished():
         pass
-    def waitUntilSuccess():
-        """Return a Deferred that fires (with this IBuildSetStatus object)
-        when the outcome of the BuildSet is known, i.e., upon the first
-        failure, or after all builds complete successfully."""
     def waitUntilFinished():
         """Return a Deferred that fires (with this IBuildSetStatus object)
         when all builds have finished."""
@@ -299,18 +234,19 @@ class IBuildRequestStatus(Interface):
     finally turned into a Build."""
 
     def getSourceStamp():
-        """Return a SourceStamp object which can be used to re-create
-        the source tree that this build used.  This method will
-        return an absolute SourceStamp if possible, and its results
-        may change as the build progresses.  Specifically, a "HEAD"
-        build may later be more accurately specified by an absolute
-        SourceStamp with the specific revision information.
+        """
+        Get a SourceStamp object which can be used to re-create the source tree
+        that this build used.  This method will return an absolute SourceStamp
+        if possible, and its results may change as the build progresses.
+        Specifically, a "HEAD" build may later be more accurately specified by
+        an absolute SourceStamp with the specific revision information.
 
         This method will return None if the source information is no longer
-        available."""
-        pass
-    def getBuilderName():
-        pass
+        available.
+
+        @returns: SourceStamp via Deferred
+        """
+
     def getBuilds():
         """Return a list of IBuildStatus objects for each Build that has been
         started in an attempt to satify this BuildRequest."""
@@ -326,9 +262,8 @@ class IBuildRequestStatus(Interface):
     def unsubscribe(observer):
         """Unregister the callable that was registered with subscribe()."""
     def getSubmitTime():
-        """Return the time when this request was submitted"""
-    def setSubmitTime(t):
-        """Sets the time when this request was submitted"""
+        """Return the time when this request was submitted.  Returns a
+        Deferred."""
 
 
 class ISlaveStatus(Interface):
@@ -365,6 +300,9 @@ class IBuilderStatus(Interface):
     def getCategory():
         """Return the category of this builder (a string)."""
 
+    def getDescription():
+        """Return the description of this builder (a string)."""
+
     def getState():
         # TODO: this isn't nearly as meaningful as it used to be
         """Return a tuple (state, builds) for this Builder. 'state' is the
@@ -377,10 +315,13 @@ class IBuilderStatus(Interface):
         """Return a list of ISlaveStatus objects for the buildslaves that are
         used by this builder."""
 
-    def getPendingBuilds():
-        """Return an IBuildRequestStatus object for all upcoming builds
-        (those which are ready to go but which are waiting for a buildslave
-        to be available."""
+    def getPendingBuildRequestStatuses():
+        """
+        Get a L{IBuildRequestStatus} implementations for all unclaimed build
+        requests.
+
+        @returns: list of objects via Deferred
+        """
 
     def getCurrentBuilds():
         """Return a list containing an IBuildStatus object for each build
@@ -508,17 +449,13 @@ class IBuildStatus(Interface):
         build has already finished, this deferred will fire right away. The
         callback is given this IBuildStatus instance as an argument."""
 
-    def getProperty(propname):
-        """Return the value of the build property with the given name. Raises
-        KeyError if there is no such property on this build."""
-
     def getReason():
         """Return a string that indicates why the build was run. 'changes',
         'forced', and 'periodic' are the most likely values. 'try' will be
         added in the future."""
 
-    def getSourceStamp():
-        """Return a SourceStamp object which can be used to re-create
+    def getSourceStamps():
+        """Return a list of SourceStamp objects which can be used to re-create
         the source tree that this build used.
 
         This method will return None if the source information is no longer
@@ -530,6 +467,10 @@ class IBuildStatus(Interface):
         """Return a list of Change objects which represent which source
         changes went into the build."""
 
+    def getRevisions():
+        """Returns a string representing the list of revisions that led to
+        the build, rendered from each Change.revision"""
+
     def getResponsibleUsers():
         """Return a list of Users who are to blame for the changes that went
         into this build. If anything breaks (at least anything that wasn't
@@ -540,10 +481,8 @@ class IBuildStatus(Interface):
 
     def getInterestedUsers():
         """Return a list of Users who will want to know about the results of
-        this build. This is a superset of getResponsibleUsers(): it adds
-        people who are interested in this build but who did not actually
-        make the Changes that went into it (build sheriffs, code-domain
-        owners)."""
+        this build but who did not actually make the Changes that went into it
+        (build sheriffs, code-domain owners)."""
 
     def getNumber():
         """Within each builder, each Build has a number. Return it."""
@@ -1024,7 +963,7 @@ class IStatusReceiver(Interface):
         in L{IBuildStatus.getResults}.
 
         @type  builderName: string
-        @type  build:       L{buildbot.status.builder.BuildStatus}
+        @type  build:       L{buildbot.status.build.BuildStatus}
         @type  results:     tuple
         """
 
@@ -1037,45 +976,42 @@ class IStatusReceiver(Interface):
     def slaveDisconnected(slaveName):
         """The slave has disconnected."""
 
+    def checkConfig(otherStatusReceivers):
+        """Verify that there are no other status receivers which conflict with
+        the current one.
+
+        @type  otherStatusReceivers: A list of L{IStatusReceiver} objects which
+        will contain self.
+        """
+
+
 class IControl(Interface):
     def addChange(change):
-        """Add a change to all builders. Each Builder will decide for
-        themselves whether the change is interesting or not, and may initiate
-        a build as a result."""
-
-    def submitBuildSet(builderNames, ss, reason, props=None, now=False):
-        """Create a BuildSet, which will eventually cause a build of the
-        given SourceStamp to be run on all of the named builders. This
-        returns a BuildSetStatus object, which can be used to keep track of
-        the builds that are performed.
-
-        If now=True, and the builder has no slave attached, NoSlaveError will
-        be raised instead of queueing the request for later action."""
+        """Add a change to the change queue, for analysis by schedulers."""
 
     def getBuilder(name):
         """Retrieve the IBuilderControl object for the given Builder."""
 
 class IBuilderControl(Interface):
-    def submitBuildRequest(ss, reason, props=None, now=False):
+    def submitBuildRequest(ss, reason, props=None):
         """Create a BuildRequest, which will eventually cause a build of the
         given SourceStamp to be run on this builder. This returns a
-        BuildRequestStatus object, which can be used to keep track of the
-        builds that are performed.
+        BuildRequestStatus object via a Deferred, which can be used to keep
+        track of the builds that are performed."""
 
-        If now=True, and I have no slave attached, NoSlaveError will be
-        raised instead of queueing the request for later action."""
-
-    def resubmitBuild(buildStatus, reason="<rebuild, no reason given>"):
+    def rebuildBuild(buildStatus, reason="<rebuild, no reason given>"):
         """Rebuild something we've already built before. This submits a
         BuildRequest to our Builder using the same SourceStamp as the earlier
         build. This has no effect (but may eventually raise an exception) if
         this Build has not yet finished."""
 
-    def getPendingBuilds():
-        """Return a list of L{IBuildRequestControl} objects for this Builder.
-        Each one corresponds to a pending build that has not yet started (due
-        to a scarcity of build slaves). These upcoming builds can be canceled
-        through the control object."""
+    def getPendingBuildRequestControls():
+        """
+        Get a list of L{IBuildRequestControl} objects for this Builder.
+        Each one corresponds to an unclaimed build request.
+
+        @returns: list of objects via Deferred
+        """
 
     def getBuild(number):
         """Attempt to return an IBuildControl object for the given build.
@@ -1167,13 +1103,114 @@ class ILatentBuildSlave(IBuildSlave):
     def buildStarted(sb):
         """Inform the latent build slave that a build has started.
 
-        ``sb`` is a LatentSlaveBuilder as defined in buildslave.py.  The sb
-        is the one for whom the build started.
+        @param sb: a L{LatentSlaveBuilder}.  The sb is the one for whom the
+        build finished.
         """
 
     def buildFinished(sb):
         """Inform the latent build slave that a build has finished.
 
-        ``sb`` is a LatentSlaveBuilder as defined in buildslave.py.  The sb
-        is the one for whom the build finished.
+        @param sb: a L{LatentSlaveBuilder}.  The sb is the one for whom the
+        build finished.
+        """
+
+class IRenderable(Interface):
+    """An object that can be interpolated with properties from a build.
+    """
+
+    def getRenderingFor(iprops):
+        """Return a deferred that fires with interpolation with the given properties
+
+        @param iprops: the L{IProperties} provider supplying the properties.
+        """
+class IProperties(Interface):
+    """
+    An object providing access to build properties
+    """
+
+    def getProperty(name, default=None):
+        """Get the named property, returning the default if the property does
+        not exist.
+
+        @param name: property name
+        @type name: string
+
+        @param default: default value (default: @code{None})
+
+        @returns: property value
+        """
+
+    def hasProperty(name):
+        """Return true if the named property exists.
+
+        @param name: property name
+        @type name: string
+        @returns: boolean
+        """
+
+    def has_key(name):
+        """Deprecated name for L{hasProperty}."""
+
+    def setProperty(name, value, source, runtime=False):
+        """Set the given property, overwriting any existing value.  The source
+        describes the source of the value for human interpretation.
+
+        @param name: property name
+        @type name: string
+
+        @param value: property value
+        @type value: JSON-able value
+
+        @param source: property source
+        @type source: string
+
+        @param runtime: (optional) whether this property was set during the
+        build's runtime: usually left at its default value
+        @type runtime: boolean
+        """
+
+    def getProperties():
+        """Get the L{buildbot.process.properties.Properties} instance storing
+        these properties.  Note that the interface for this class is not
+        stable, so where possible the other methods of this interface should be
+        used.
+
+        @returns: L{buildbot.process.properties.Properties} instance
+        """
+
+    def getBuild():
+        """Get the L{buildbot.process.build.Build} instance for the current
+        build.  Note that this object is not available after the build is
+        complete, at which point this method will return None.
+
+        Try to avoid using this method, as the API of L{Build} instances is not
+        well-defined.
+
+        @returns L{buildbot.process.build.Build} instance
+        """
+
+    def render(value):
+        """Render @code{value} as an L{IRenderable}.  This essentially coerces
+        @code{value} to an L{IRenderable} and calls its @L{getRenderingFor}
+        method.
+
+        @name value: value to render
+        @returns: rendered value
+        """
+
+class IScheduler(Interface):
+    pass
+
+class ITriggerableScheduler(Interface):
+    """
+    A scheduler that can be triggered by buildsteps.
+    """
+
+    def trigger(sourcestamps, set_props=None):
+        """Trigger a build with the given source stamp and properties.
+        """
+
+class IBuildStepFactory(Interface):
+    def buildStep():
+        """
         """
