@@ -1,22 +1,38 @@
+# This file is part of Buildbot.  Buildbot is free software: you can
+# redistribute it and/or modify it under the terms of the GNU General Public
+# License as published by the Free Software Foundation, version 2.
+#
+# This program is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
+# details.
+#
+# You should have received a copy of the GNU General Public License along with
+# this program; if not, write to the Free Software Foundation, Inc., 51
+# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+#
+# Copyright Buildbot Team Members
+
 
 import os, weakref
 
 from zope.interface import implements
 from twisted.python import log
 from twisted.application import strports, service
+from twisted.internet import defer
 from twisted.web import server, distrib, static
 from twisted.spread import pb
 from twisted.web.util import Redirect
-
+from buildbot import config
 from buildbot.interfaces import IStatusReceiver
-
 from buildbot.status.web.base import StaticFile, createJinjaEnv
 from buildbot.status.web.feeds import Rss20StatusResource, \
      Atom10StatusResource
 from buildbot.status.web.waterfall import WaterfallStatusResource
 from buildbot.status.web.console import ConsoleStatusResource
 from buildbot.status.web.olpb import OneLinePerBuild
-from buildbot.status.web.grid import GridStatusResource, TransposedGridStatusResource
+from buildbot.status.web.grid import GridStatusResource
+from buildbot.status.web.grid import TransposedGridStatusResource
 from buildbot.status.web.changes import ChangesResource
 from buildbot.status.web.builder import BuildersResource
 from buildbot.status.web.buildstatus import BuildStatusStatusResource
@@ -24,9 +40,15 @@ from buildbot.status.web.slaves import BuildSlavesResource
 from buildbot.status.web.status_json import JsonStatusResource
 from buildbot.status.web.about import AboutBuildbot
 from buildbot.status.web.authz import Authz
-from buildbot.status.web.auth import AuthFailResource
+from buildbot.status.web.auth import AuthFailResource,AuthzFailResource, LoginResource, LogoutResource
 from buildbot.status.web.root import RootPage
+from buildbot.status.web.users import UsersResource
 from buildbot.status.web.change_hook import ChangeHookResource
+from twisted.cred.portal import IRealm, Portal
+from twisted.cred import strcred
+from twisted.cred.checkers import ICredentialsChecker
+from twisted.cred.credentials import IUsernamePassword
+from twisted.web import resource, guard
 
 # this class contains the WebStatus class.  Basic utilities are in base.py,
 # and specific pages are each in their own module.
@@ -97,7 +119,7 @@ class WebStatus(service.MultiService):
     generates the root page.
 
     All of the resources provided by this service use relative URLs to reach
-    each other. The only absolute links are the c['projectURL'] links at the
+    each other. The only absolute links are the c['titleURL'] links at the
     top and bottom of the page, and the buildbot home-page link at the
     bottom.
 
@@ -132,7 +154,8 @@ class WebStatus(service.MultiService):
                  order_console_by_time=False, changecommentlink=None,
                  revlink=None, projects=None, repositories=None,
                  authz=None, logRotateLength=None, maxRotatedFiles=None,
-                 change_hook_dialects = {}):
+                 change_hook_dialects = {}, provide_feeds=None, jinja_loaders=None,
+                 change_hook_auth=None):
         """Run a web server that provides Buildbot status.
 
         @type  http_port: int or L{twisted.application.strports} string
@@ -245,6 +268,15 @@ class WebStatus(service.MultiService):
                                      
         
     
+        @type  provide_feeds: None or list
+        @param provide_feeds: If empty, provides atom, json, and rss feeds.
+                              Otherwise, a dictionary of strings of
+                              the type of feeds provided.  Current
+                              possibilities are "atom", "json", and "rss"
+
+        @type  jinja_loaders: None or list
+        @param jinja_loaders: If not empty, a list of additional Jinja2 loader
+                              objects to search for templates.
         """
 
         service.MultiService.__init__(self)
@@ -259,17 +291,21 @@ class WebStatus(service.MultiService):
         self.distrib_port = distrib_port
         self.num_events = num_events
         if num_events_max:
-            assert num_events_max >= num_events
+            if num_events_max < num_events:
+                config.error(
+                    "num_events_max must be greater than num_events")
             self.num_events_max = num_events_max
         self.public_html = public_html
 
         # make up an authz if allowForce was given
         if authz:
             if allowForce is not None:
-                raise ValueError("cannot use both allowForce and authz parameters")
+                config.error(
+                    "cannot use both allowForce and authz parameters")
             if auth:
-                raise ValueError("cannot use both auth and authz parameters (pass "
-                                 "auth as an Authz parameter)")
+                config.error(
+                    "cannot use both auth and authz parameters (pass " +
+                    "auth as an Authz parameter)")
         else:
             # invent an authz
             if allowForce and auth:
@@ -284,11 +320,37 @@ class WebStatus(service.MultiService):
 
         self.authz = authz
 
+        # check for correctness of HTTP auth parameters
+        if change_hook_auth is not None:
+            self.change_hook_auth = []
+            for checker in change_hook_auth:
+                if isinstance(checker, str):
+                    try:
+                        checker = strcred.makeChecker(checker)
+                    except Exception, error:
+                        config.error("Invalid change_hook checker description: %s" % (error,))
+                        continue
+                elif not ICredentialsChecker.providedBy(checker):
+                    config.error("change_hook checker doesn't provide ICredentialChecker: %r" % (checker,))
+                    continue
+
+                if IUsernamePassword not in checker.credentialInterfaces:
+                    config.error("change_hook checker doesn't support IUsernamePassword: %r" % (checker,))
+                    continue
+
+                self.change_hook_auth.append(checker)
+        else:
+            self.change_hook_auth = None
+
         self.orderConsoleByTime = order_console_by_time
 
         # If we were given a site object, go ahead and use it. (if not, we add one later)
         self.site = site
 
+        # keep track of our child services
+        self.http_svc = None
+        self.distrib_svc = None
+
         # store the log settings until we create the site object
         self.logRotateLength = logRotateLength
         self.maxRotatedFiles = maxRotatedFiles        
@@ -298,9 +360,10 @@ class WebStatus(service.MultiService):
         self.setupUsualPages(numbuilds=numbuilds, num_events=num_events,
                              num_events_max=num_events_max)
 
-        # Set up the jinja templating engine.
-        self.templates = createJinjaEnv(revlink, changecommentlink,
-                                        repositories, projects)
+        self.revlink = revlink
+        self.changecommentlink = changecommentlink
+        self.repositories = repositories
+        self.projects = projects
 
         # keep track of cached connections so we can break them when we shut
         # down. See ticket #102 for more details.
@@ -310,7 +373,37 @@ class WebStatus(service.MultiService):
         self.change_hook_dialects = {}
         if change_hook_dialects:
             self.change_hook_dialects = change_hook_dialects
-            self.putChild("change_hook", ChangeHookResource(dialects = self.change_hook_dialects))
+            resource_obj = ChangeHookResource(dialects=self.change_hook_dialects)
+            if self.change_hook_auth is not None:
+                resource_obj = self.setupProtectedResource(
+                        resource_obj, self.change_hook_auth)
+            self.putChild("change_hook", resource_obj)
+
+        # Set default feeds
+        if provide_feeds is None:
+            self.provide_feeds = ["atom", "json", "rss"]
+        else:
+            self.provide_feeds = provide_feeds
+
+        self.jinja_loaders = jinja_loaders
+
+    def setupProtectedResource(self, resource_obj, checkers):
+        class SimpleRealm(object):
+            """
+            A realm which gives out L{ChangeHookResource} instances for authenticated
+            users.
+            """
+            implements(IRealm)
+
+            def requestAvatar(self, avatarId, mind, *interfaces):
+                if resource.IResource in interfaces:
+                    return (resource.IResource, resource_obj, lambda: None)
+                raise NotImplementedError()
+
+        portal = Portal(SimpleRealm(), checkers)
+        credentialFactory = guard.BasicCredentialFactory('Protected area')
+        wrapper = guard.HTTPAuthSessionWrapper(portal, [credentialFactory])
+        return wrapper
 
     def setupUsualPages(self, numbuilds, num_events, num_events_max):
         #self.putChild("", IndexOrWaterfallRedirection())
@@ -320,7 +413,7 @@ class WebStatus(service.MultiService):
         self.putChild("console", ConsoleStatusResource(
                 orderByTime=self.orderConsoleByTime))
         self.putChild("tgrid", TransposedGridStatusResource())
-        self.putChild("builders", BuildersResource()) # has builds/steps/logs
+        self.putChild("builders", BuildersResource(numbuilds=numbuilds)) # has builds/steps/logs
         self.putChild("one_box_per_builder", Redirect("builders"))
         self.putChild("changes", ChangesResource())
         self.putChild("buildslaves", BuildSlavesResource())
@@ -329,7 +422,10 @@ class WebStatus(service.MultiService):
                       OneLinePerBuild(numbuilds=numbuilds))
         self.putChild("about", AboutBuildbot())
         self.putChild("authfail", AuthFailResource())
-
+        self.putChild("authzfail", AuthzFailResource())
+        self.putChild("users", UsersResource())
+        self.putChild("login", LoginResource())
+        self.putChild("logout", LogoutResource())
 
     def __repr__(self):
         if self.http_port is None:
@@ -342,14 +438,16 @@ class WebStatus(service.MultiService):
                 (self.http_port, self.distrib_port, hex(id(self))))
 
     def setServiceParent(self, parent):
-        service.MultiService.setServiceParent(self, parent)
-
         # this class keeps a *separate* link to the buildmaster, rather than
         # just using self.parent, so that when we are "disowned" (and thus
         # parent=None), any remaining HTTP clients of this WebStatus will still
         # be able to get reasonable results.
-        self.master = parent
-        
+        self.master = parent.master
+
+        # set master in IAuth instance
+        if self.authz.auth:
+            self.authz.auth.master = self.master
+
         def either(a,b): # a if a else b for py2.4
             if a:
                 return a
@@ -359,6 +457,14 @@ class WebStatus(service.MultiService):
         rotateLength = either(self.logRotateLength, self.master.log_rotation.rotateLength)
         maxRotatedFiles = either(self.maxRotatedFiles, self.master.log_rotation.maxRotatedFiles)
 
+        # Set up the jinja templating engine.
+        if self.revlink:
+            revlink = self.revlink
+        else:
+            revlink = self.master.config.revlink
+        self.templates = createJinjaEnv(revlink, self.changecommentlink,
+                                        self.repositories, self.projects, self.jinja_loaders)
+
         if not self.site:
             
             class RotateLogSite(server.Site):
@@ -388,15 +494,17 @@ class WebStatus(service.MultiService):
         self.site.buildbot_service = self
 
         if self.http_port is not None:
-            s = strports.service(self.http_port, self.site)
+            self.http_svc = s = strports.service(self.http_port, self.site)
             s.setServiceParent(self)
         if self.distrib_port is not None:
             f = pb.PBServerFactory(distrib.ResourcePublisher(self.site))
-            s = strports.service(self.distrib_port, f)
+            self.distrib_svc = s = strports.service(self.distrib_port, f)
             s.setServiceParent(self)
 
         self.setupSite()
 
+        service.MultiService.setServiceParent(self, parent)
+
     def setupSite(self):
         # this is responsible for creating the root resource. It isn't done
         # at __init__ time because we need to reference the parent's basedir.
@@ -421,9 +529,12 @@ class WebStatus(service.MultiService):
             root.putChild(name, child_resource)
 
         status = self.getStatus()
-        root.putChild("rss", Rss20StatusResource(status))
-        root.putChild("atom", Atom10StatusResource(status))
-        root.putChild("json", JsonStatusResource(status))
+        if "rss" in self.provide_feeds:
+            root.putChild("rss", Rss20StatusResource(status))
+        if "atom" in self.provide_feeds:
+            root.putChild("atom", Atom10StatusResource(status))
+        if "json" in self.provide_feeds:
+            root.putChild("json", JsonStatusResource(status))
 
         self.site.resource = root
 
@@ -434,6 +545,7 @@ class WebStatus(service.MultiService):
     def registerChannel(self, channel):
         self.channels[channel] = 1 # weakrefs
 
+    @defer.inlineCallbacks
     def stopService(self):
         for channel in self.channels:
             try:
@@ -442,7 +554,16 @@ class WebStatus(service.MultiService):
                 log.msg("WebStatus.stopService: error while disconnecting"
                         " leftover clients")
                 log.err()
-        return service.MultiService.stopService(self)
+        yield service.MultiService.stopService(self)
+
+        # having shut them down, now remove our child services so they don't
+        # start up again if we're re-started
+        if self.http_svc:
+            yield self.http_svc.disownServiceParent()
+            self.http_svc = None
+        if self.distrib_svc:
+            yield self.distrib_svc.disownServiceParent()
+            self.distrib_svc = None
 
     def getStatus(self):
         return self.master.getStatus()
@@ -463,6 +584,24 @@ class WebStatus(service.MultiService):
     # This is in preparation for removal of the IControl hierarchy
     # entirely.
 
+    def checkConfig(self, otherStatusReceivers):
+        duplicate_webstatus=0
+        for osr in otherStatusReceivers:
+            if isinstance(osr,WebStatus):
+                if osr is self:
+                    continue
+                # compare against myself and complain if the settings conflict
+                if self.http_port == osr.http_port:
+                    if duplicate_webstatus == 0:
+                        duplicate_webstatus = 2
+                    else:
+                        duplicate_webstatus += 1
+
+        if duplicate_webstatus:
+            config.error(
+                "%d Webstatus objects have same port: %s"
+                    % (duplicate_webstatus, self.http_port),
+            )
+
 # resources can get access to the IStatus by calling
 # request.site.buildbot_service.getStatus()
-
